#!/usr/bin/env python3
"""
__main__.py for Freqtrade
To launch Freqtrade as a module

> python -m freqtrade (with Python >= 3.9)
"""

from freqtrade import main


if __name__ == "__main__":
    main.main()
